#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <unistd.h>
#include <signal.h>
#include <inttypes.h>
#include <time.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/utsname.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <sys/mman.h>
#include <netinet/in.h>

#define xc_evtchn_open              xenner_evtchn_open
#define xc_evtchn_fd                xenner_evtchn_fd
#define xc_evtchn_bind_unbound_port xenner_evtchn_bind_unbound_port
#define xc_evtchn_bind_interdomain  xenner_evtchn_bind_interdomain
#define xc_evtchn_bind_virq         xenner_evtchn_bind_virq
#define xc_evtchn_unbind            xenner_evtchn_unbind
#define xc_evtchn_notify            xenner_evtchn_notify
#define xc_evtchn_pending           xenner_evtchn_pending
#define xc_evtchn_unmask            xenner_evtchn_unmask
#define xc_evtchn_close             xenner_evtchn_close

#include <xenctrl.h>
#include <xs.h>
#include <xen/io/xenbus.h>

#include <xen/elfnote.h>
#include <xen/elfstructs.h>

#include "blkif.h"
#include "shared.h"
#include "list.h"
#include "hexdump.h"
#include "daemon.h"

/* ----------------------------------------------------------- */

static int libxc_trace;
static int libxc_fixme;

#include "evtchnd.h"
#include "libxc-evtchn.c"

/* ------------------------------------------------------------- */

#ifndef offsetof
#define offsetof(TYPE, MEMBER) ((size_t) &((TYPE *)0)->MEMBER)
#endif

#define container_of(ptr, type, member) ({                      \
        const typeof( ((type *)0)->member ) *__mptr = (ptr);    \
        (type *)( (char *)__mptr - offsetof(type,member) );})

/* ------------------------------------------------------------- */

#define BUFSIZE 256
#define DEVPATH "/local/domain/0/backend/blkbackd"

struct xendev;

struct devops {
    size_t size;
    int    (*alloc)(struct xendev *xendev);
    int    (*setup)(struct xendev *xendev);
    int    (*xs_be)(struct xendev *xendev, char *node, char *val);
    int    (*xs_fe)(struct xendev *xendev, char *node, char *val);
    int    (*event)(struct xendev *xendev);
    int    (*free)(struct xendev *xendev);
};

struct xendev {
    int                dom;
    int                dev;
    enum xenbus_state  be_state;
    enum xenbus_state  fe_state;
    char               *fe;
    char               *protocol;
    int                ring_ref;
    int                remote_port;
    int                local_port;

    void               *vmcore;
    size_t             vmcore_size;
    Elf64_Ehdr         *vm_ehdr;
    Elf64_Phdr         *vm_phdrs;
    Elf64_Shdr         *vm_shdrs;
    char               *vm_shstrtab;
    struct grant_entry *grant_table;

    struct xs_handle   *xs;
    struct devops      *ops;
    struct list_head   next;
};
static LIST_HEAD(xendevs);
static int devcount;

struct blkdev {
    struct xendev       xendev;  /* must be first */
    char                *params;
    char                *mode;
    int                 file;
    size_t              file_blk;
    size_t              file_size;
    int                 protocol;
    blkif_back_rings_t  rings;
    int                 more_work;
};

static int evtchnd;
static int termsig;

/* ------------------------------------------------------------- */

static void catchsig(int sig)
{
    termsig = sig;
}

/* ------------------------------------------------------------- */

static struct xendev *find_xendev(int dom, int dev)
{
    struct xendev *xendev;
    struct list_head *item;

    list_for_each(item, &xendevs) {
	xendev = list_entry(item, struct xendev, next);
	if (xendev->dom == dom  &&  xendev->dev == dev)
	    return xendev;
    }
    return NULL;
}

static struct xendev *get_xendev(int dom, int dev,
				 struct xs_handle *xs,
				 struct devops *ops)
{
    struct xendev *xendev;

    xendev = find_xendev(dom, dev);
    if (!xendev) {
	xendev = malloc(ops->size);
	memset(xendev,0,ops->size);
	xendev->dom = dom;
	xendev->dev = dev;
	xendev->xs  = xs;
	xendev->ops = ops;

	list_add_tail(&xendev->next, &xendevs);
	devcount++;
	d1printf("%s: %d/%d\n", __FUNCTION__,
		 xendev->dom, xendev->dev);

	if (xendev->ops->alloc)
	    xendev->ops->alloc(xendev);
    }
    return xendev;
}

static struct xendev *del_xendev(int dom, int dev)
{
    struct xendev *xendev;
    struct list_head *item, *tmp;

    list_for_each_safe(item, tmp, &xendevs) {
	xendev = list_entry(item, struct xendev, next);
	if (xendev->dom != dom)
	    continue;
	if (xendev->dev != dev && dev != -1)
	    continue;

	if (xendev->ops->free)
	    xendev->ops->free(xendev);

	if (xendev->fe) {
	    char token[BUFSIZE];
	    snprintf(token, sizeof(token), "fe:%p", xendev);
	    xs_unwatch(xendev->xs, xendev->fe, token);
	    free(xendev->fe);
	}
	
	d1printf("%s: %d/%d\n", __FUNCTION__,
		 xendev->dom, xendev->dev);
	list_del(&xendev->next);
	devcount--;
	free(xendev);
    }
    return NULL;
}

/* ------------------------------------------------------------- */

static int write_be_str(struct xendev *xendev, char *node, char *val)
{
    char abspath[BUFSIZE];
    
    snprintf(abspath, sizeof(abspath), DEVPATH "/%d/%d/%s",
	     xendev->dom, xendev->dev, node);
    return xs_write(xendev->xs, 0, abspath, val, strlen(val));
}

static int write_be_int(struct xendev *xendev, char *node, int ival)
{
    char val[32];

    snprintf(val, sizeof(val), "%d", ival);
    return write_be_str(xendev, node, val);
}

static char *read_be_str(struct xendev *xendev, char *node)
{
    char abspath[BUFSIZE];
    unsigned int len;

    snprintf(abspath, sizeof(abspath), DEVPATH "/%d/%d/%s",
	     xendev->dom, xendev->dev, node);
    return xs_read(xendev->xs, 0, abspath, &len);
}

static int read_be_int(struct xendev *xendev, char *node)
{
    char *val;
    int ival;

    val = read_be_str(xendev, node);
    ival = val ? atoi(val) : 0;
    free(val);
    return ival;
}

static char *read_fe_str(struct xendev *xendev, char *node)
{
    char abspath[BUFSIZE];
    unsigned int len;

    snprintf(abspath, sizeof(abspath), "%s/%s",
	     xendev->fe, node);
    return xs_read(xendev->xs, 0, abspath, &len);
}

static int read_fe_int(struct xendev *xendev, char *node)
{
    char *val;
    int ival;

    val = read_fe_str(xendev, node);
    ival = val ? atoi(val) : 0;
    free(val);
    return ival;
}

static int check_state_xendev(struct xendev *xendev, int be, int fe)
{
    enum xenbus_state state;

    if (be) {
	state = read_be_int(xendev, "state");
	if (xendev->be_state != state) {
	    d1printf("%s: be: %d -> %d\n", __FUNCTION__,
		     xendev->be_state, state);
	    xendev->be_state  = state;
	}
    }

    if (fe) {
	state = read_fe_int(xendev, "state");
	if (xendev->fe_state != state) {
	    d1printf("%s: fe: %d -> %d\n", __FUNCTION__,
		     xendev->fe_state, state);
	    xendev->fe_state  = state;
	}
    }
    return 0;
}

static int change_state_xendev(struct xendev *xendev, enum xenbus_state state)
{
    write_be_int(xendev, "state", state);
    d1printf("%s: be: %d -> %d\n", __FUNCTION__,
	     xendev->be_state, state);
    xendev->be_state  = state;
    return 0;
}

static int setup_xendev(struct xendev *xendev)
{
    char token[BUFSIZE];

    if (NULL != xendev->fe)
	/* already done */
	return 0;

    xendev->fe = read_be_str(xendev, "frontend");
    if (NULL == xendev->fe)
	/* no fe ptr yet */
	return 0;

    /* setup frontend watch */
    snprintf(token, sizeof(token), "fe:%p", xendev);
    xs_watch(xendev->xs, xendev->fe, token);

    check_state_xendev(xendev, 1, 1);

    if (xendev->ops->setup)
	xendev->ops->setup(xendev);
    return 0;
}

/* ------------------------------------------------------------- */

static int vmcore_map(struct xendev *xendev)
{
    char filename[256];
    struct stat st;
    char *name;
    int i, fd;

    snprintf(filename, sizeof(filename), "/var/run/xenner/vmcore.%d",
	     xendev->dom);
    fd = open(filename, O_RDWR);
    if (-1 == fd)
	return -1;
    fstat(fd, &st);
    xendev->vmcore_size = st.st_size;
    xendev->vmcore = mmap(NULL, xendev->vmcore_size,
			  PROT_READ | PROT_WRITE, MAP_SHARED, fd, 0);
    if ((void*)-1 == xendev->vmcore) {
	d1printf("%s: mmap(%s,%zd): %s\n",
		 __FUNCTION__, filename, xendev->vmcore_size, strerror(errno));
	xendev->vmcore = NULL;
	close(fd);
	return -1;
    }
    close(fd);

    xendev->vm_ehdr  = xendev->vmcore;
    xendev->vm_phdrs = xendev->vmcore + xendev->vm_ehdr->e_phoff;
    xendev->vm_shdrs = xendev->vmcore + xendev->vm_ehdr->e_shoff;
    xendev->vm_shstrtab = xendev->vmcore +
	xendev->vm_shdrs[xendev->vm_ehdr->e_shstrndx].sh_offset;

    for (i = 0; i < xendev->vm_ehdr->e_shnum; i++) {
	name = xendev->vm_shstrtab + xendev->vm_shdrs[i].sh_name;
	d1printf("  %d: %s\n", i, name);
	if (0 == strcmp(name, ".addr.grant_table"))
	    xendev->grant_table = xendev->vmcore + xendev->vm_shdrs[i].sh_offset;
    }
	
    return 0;
}

static int vmcore_unmap(struct xendev *xendev)
{
    munmap(xendev->vmcore, xendev->vmcore_size);
    return 0;
}

static void *vmcore_mfn2ptr(struct xendev *xendev, uint64_t mfn)
{
    uint64_t offset;
    void *ptr;

    offset = VMCORE_HEADER + mfn * PAGE_SIZE;
    if (offset >= xendev->vmcore_size)
	return NULL;
    ptr = xendev->vmcore + offset;
    d2printf("%s: mfn 0x%" PRIx64 " -> %p\n", __FUNCTION__, mfn, ptr);
    return ptr;
}

static void *vmcore_ref2ptr(struct xendev *xendev, uint32_t ref)
{
    struct grant_entry *e;
    
    if (!xendev->grant_table)
	return NULL;

    e = xendev->grant_table + ref;
    if ((e->flags & GTF_type_mask) != GTF_permit_access)
	goto err;
    
    return vmcore_mfn2ptr(xendev, e->frame);

err:
    d1printf("%s: flags 0x%x, domid %d, frame 0x%x FAILED\n", __FUNCTION__,
	     (int)xendev->grant_table[ref].flags,
	     (int)xendev->grant_table[ref].domid,
	     (int)xendev->grant_table[ref].frame);
    return NULL;
}

/* ------------------------------------------------------------- */

static void xenstore_scan(struct xs_handle *xs, char *devpath,
			  struct devops *ops)
{
    struct xendev *xendev;
    char path[BUFSIZE], token[BUFSIZE];
    char **dom = NULL;
    char **dev = NULL;
    unsigned int cdom = 0, cdev = 0, i, j;
        
    /* setup watch */
    snprintf(token, sizeof(token), "be:%p", ops);
    xs_watch(xs, DEVPATH, token);

    /* look for backends */
    d1printf("%s: %s\n", __FUNCTION__, devpath);
    dom = xs_directory(xs, 0, devpath, &cdom);
    for (i = 0; i < cdom; i++) {
	d1printf("%s:   %s\n", __FUNCTION__, dom[i]);
	snprintf(path, sizeof(path), "%s/%s", DEVPATH, dom[i]);
	dev = xs_directory(xs, 0, path, &cdev);
	for (j = 0; j < cdev; j++) {
	    d1printf("%s:     %s\n", __FUNCTION__, dev[j]);
	    xendev = get_xendev(atoi(dom[i]), atoi(dev[j]), xs, ops);
	    setup_xendev(xendev);
	}
	if (dev)
	    free(dev);
    }
    if (dom)
	free(dom);
}

static void xenstore_update_be(struct xs_handle *xs, char *watch, struct devops *ops)
{
    struct xendev *xendev;
    char path[BUFSIZE], *val;
    int dom, dev;
    unsigned int len;
        
    if (3 != sscanf(watch, DEVPATH "/%d/%d/%255s", &dom, &dev, path)) {
	strcpy(path, "");
	if (2 != sscanf(watch, DEVPATH "/%d/%d", &dom, &dev)) {
	    dev = -1;
	    if (1 != sscanf(watch, DEVPATH "/%d", &dom))
		return;
	}
    }
    val = xs_read(xs, 0, watch, &len);

    if (NULL == val) {
	del_xendev(dom, dev);
	return;
    }

    xendev = get_xendev(dom, dev, xs, ops);
    setup_xendev(xendev);
    if (xendev->fe && xendev->ops->xs_be)
	xendev->ops->xs_be(xendev, path, val);

    if (val)
	free(val);
}

static void xenstore_update_fe(struct xs_handle *xs, char *watch, struct xendev *xendev)
{
    char *val, *node;
    unsigned int len;

    val = xs_read(xs, 0, watch, &len);

    node = strrchr(watch, '/');
    if (node) {
	node++;
	if (0 == strcmp(node, "state"))
	    check_state_xendev(xendev, 0, 1);
    }
    if (xendev->ops->xs_fe)
	xendev->ops->xs_fe(xendev, node, val);

    if (val)
	free(val);
}

static void xenstore_update(struct xs_handle *xs)
{
    char **vec = NULL;
    intptr_t ptr;
    unsigned int count;
        
    vec = xs_read_watch(xs, &count);
    if (NULL == vec)
	goto cleanup;
    d2printf("%s: %s [%s]\n", __FUNCTION__,
	     vec[XS_WATCH_PATH], vec[XS_WATCH_TOKEN]);

    if (1 == sscanf(vec[XS_WATCH_TOKEN], "be:%" PRIxPTR, &ptr))
	xenstore_update_be(xs, vec[XS_WATCH_PATH], (void*)ptr);
    if (1 == sscanf(vec[XS_WATCH_TOKEN], "fe:%" PRIxPTR, &ptr))
	xenstore_update_fe(xs, vec[XS_WATCH_PATH], (void*)ptr);

cleanup:
    if (vec)
	free(vec);
}

static void event_received(void)
{
    struct xendev *xendev;
    struct list_head *item;
    int port;
    
    port = xenner_evtchn_pending(evtchnd);
    if (-1 == port) {
	d1printf("%s: evtchnd failure, huh?\n", __FUNCTION__);
	exit(1);
    }

    list_for_each(item, &xendevs) {
	xendev = list_entry(item, struct xendev, next);
	if (xendev->local_port != port)
	    continue;
	d2printf("%s: port %d -> dom %d, dev %d\n", __FUNCTION__,
		 port, xendev->dom, xendev->dev);
	if (xendev->ops->event)
	    xendev->ops->event(xendev);
	return;
    }
    d1printf("%s: port %d -> Huh, not found?\n", __FUNCTION__, port);
}

/* ------------------------------------------------------------- */

static void blk_setup_backend(struct blkdev *blkdev)
{
    struct stat st;
    
    /* read xenstore entries */
    if (NULL == blkdev->params)
	blkdev->params = read_be_str(&blkdev->xendev, "params");
    if (NULL == blkdev->mode)
	blkdev->mode = read_be_str(&blkdev->xendev, "mode");

    /* do we have all we need? */
    if (NULL == blkdev->params ||
	NULL == blkdev->mode)
	return;

    if (-1 == blkdev->file) {
	/* try opening file */
	int mode = O_RDONLY;
	if (0 == strcmp(blkdev->mode, "w"))
	    mode = O_RDWR;
	blkdev->file = open(blkdev->params, mode);
	if (-1 == blkdev->file) {
	    d1printf("%s: open %s: %s\n", __FUNCTION__,
		     blkdev->params, strerror(errno));
	    return;
	}

	/* fill info */
	fstat(blkdev->file, &st);
	blkdev->file_blk  = 512;
	blkdev->file_size = st.st_size;
	write_be_int(&blkdev->xendev, "feature-barrier", 1);
	write_be_int(&blkdev->xendev, "info",            0);
	write_be_int(&blkdev->xendev, "sector-size",     blkdev->file_blk);
	write_be_int(&blkdev->xendev, "sectors",
		     blkdev->file_size / blkdev->file_blk);
	change_state_xendev(&blkdev->xendev, XenbusStateInitWait);
    }
}

static void blk_connect_frontend(struct blkdev *blkdev)
{
    void *sring;
    
    if (-1 == blkdev->file)
	return; /* not ready yet */

    if (0 != blkdev->xendev.local_port)
	return; /* already done */

    blkdev->xendev.ring_ref = read_fe_int(&blkdev->xendev, "ring-ref");
    blkdev->xendev.remote_port = read_fe_int(&blkdev->xendev, "event-channel");
    blkdev->xendev.protocol = read_fe_str(&blkdev->xendev, "protocol");
    if (0 == blkdev->xendev.ring_ref ||
	0 == blkdev->xendev.remote_port ||
	NULL == blkdev->xendev.protocol)
	return;

    blkdev->protocol = BLKIF_PROTOCOL_NATIVE;
    if (0 == strcmp(blkdev->xendev.protocol, XEN_IO_PROTO_ABI_X86_32))
	blkdev->protocol = BLKIF_PROTOCOL_X86_32;
    if (0 == strcmp(blkdev->xendev.protocol, XEN_IO_PROTO_ABI_X86_64))
	blkdev->protocol = BLKIF_PROTOCOL_X86_64;

    if (vmcore_map(&blkdev->xendev) < 0)
	return;
    sring = vmcore_ref2ptr(&blkdev->xendev, blkdev->xendev.ring_ref);
    if (!sring)
	return;
    switch (blkdev->protocol) {
    case BLKIF_PROTOCOL_NATIVE:
    {
	blkif_sring_t *sring_native = sring;
	BACK_RING_INIT(&blkdev->rings.native, sring_native, PAGE_SIZE);
	break;
    }
    case BLKIF_PROTOCOL_X86_32:
    {
	blkif_x86_32_sring_t *sring_x86_32 = sring;
	BACK_RING_INIT(&blkdev->rings.x86_32, sring_x86_32, PAGE_SIZE);
	break;
    }
    case BLKIF_PROTOCOL_X86_64:
    {
	blkif_x86_64_sring_t *sring_x86_64 = sring;
	BACK_RING_INIT(&blkdev->rings.x86_64, sring_x86_64, PAGE_SIZE);
	break;
    }
    }

    blkdev->xendev.local_port = xenner_evtchn_bind_interdomain
	(evtchnd, blkdev->xendev.dom, blkdev->xendev.remote_port);

    d1printf("%s: ok: proto %s, ring-ref %d, "
	     "remote port %d, local port %d\n", __FUNCTION__,
	     blkdev->xendev.protocol, blkdev->xendev.ring_ref,
	     blkdev->xendev.remote_port, blkdev->xendev.local_port);
    change_state_xendev(&blkdev->xendev, XenbusStateConnected);
}

static void blk_send_response(struct blkdev *blkdev, uint64_t id,
			      unsigned short op, int st)
{
    int               send_notify   = 0;
    int               have_requests = 0;
    blkif_response_t  resp;
    void              *dst;
    
    resp.id        = id;
    resp.operation = op;
    resp.status    = st;
    
    /* Place on the response ring for the relevant domain. */
    switch (blkdev->protocol) {
    case BLKIF_PROTOCOL_NATIVE:
	dst = RING_GET_RESPONSE(&blkdev->rings.native, blkdev->rings.native.rsp_prod_pvt);
	break;
    case BLKIF_PROTOCOL_X86_32:
	dst = RING_GET_RESPONSE(&blkdev->rings.x86_32, blkdev->rings.x86_32.rsp_prod_pvt);
	break;
    case BLKIF_PROTOCOL_X86_64:
	dst = RING_GET_RESPONSE(&blkdev->rings.x86_64, blkdev->rings.x86_64.rsp_prod_pvt);
	break;
    }
    memcpy(dst, &resp, sizeof(resp));
    blkdev->rings.common.rsp_prod_pvt++;

    RING_PUSH_RESPONSES_AND_CHECK_NOTIFY(&blkdev->rings.common, send_notify);
    if (blkdev->rings.common.rsp_prod_pvt == blkdev->rings.common.req_cons) {
	/*
	 * Tail check for pending requests. Allows frontend to avoid
	 * notifications if requests are already in flight (lower
	 * overheads and promotes batching).
	 */
	RING_FINAL_CHECK_FOR_REQUESTS(&blkdev->rings.common, have_requests);
    } else if (RING_HAS_UNCONSUMED_REQUESTS(&blkdev->rings.common)) {
	have_requests = 1;
    }

    d2printf("%s: send_notify %d, have_requests %d\n",
	     __FUNCTION__, send_notify, have_requests);

    if (have_requests)
	blkdev->more_work++;
    if (send_notify)
	xenner_evtchn_notify(evtchnd, blkdev->xendev.local_port);
}

static void blk_handle_request(struct blkdev *blkdev, blkif_request_t *req)
{
    struct iovec vec[BLKIF_MAX_SEGMENTS_PER_REQUEST];
    void *page, *mem;
    off_t start, end;
    size_t len;
    int i;
    
    d2printf("%s: op %d, nr %d, handle %d, id %" PRId64 ", sector %" PRId64 "\n",
	     __FUNCTION__, req->operation, req->nr_segments,
	     req->handle, req->id, req->sector_number);
    if (BLKIF_OP_READ != req->operation && blkdev->mode[0] != 'w')
	goto err; /* write request for r/o disk */

    start = end = req->sector_number * blkdev->file_blk;
    for (i = 0; i < req->nr_segments; i++) {
	d2printf("%s:   %d: gref %d sects %d-%d\n", __FUNCTION__,
		 i, req->seg[i].gref, req->seg[i].first_sect, req->seg[i].last_sect);

	page = vmcore_ref2ptr(&blkdev->xendev, req->seg[i].gref);
	if (NULL == page)
	    goto err;
	len = (req->seg[i].last_sect - req->seg[i].first_sect + 1) * blkdev->file_blk;
	mem = page + req->seg[i].first_sect * blkdev->file_blk;
	d1printf("%s:   %d: op %d, mem %p, len %zd\n",
		 __FUNCTION__, i, req->operation, mem, len);
	vec[i].iov_base = mem;
	vec[i].iov_len  = len;
	end += len;
    }
    if (end >= blkdev->file_size)
	goto err; /* access beyoind end of file */

    /* run i/o */
    lseek(blkdev->file, start, SEEK_SET);
    switch (req->operation) {
    case BLKIF_OP_READ:
	len = readv(blkdev->file, vec, req->nr_segments);
	break;
    case BLKIF_OP_WRITE_BARRIER:
	fdatasync(blkdev->file);
	len = writev(blkdev->file, vec, req->nr_segments);
	fdatasync(blkdev->file);
	break;
    case BLKIF_OP_WRITE:
	len = writev(blkdev->file, vec, req->nr_segments);
	break;
    default:
	goto err; /* unknown operation */
    }
    if (len != end - start)
	goto err; /* I/O error */

    blk_send_response(blkdev, req->id, req->operation, BLKIF_RSP_OKAY);
    return;

err:
    blk_send_response(blkdev, req->id, req->operation, BLKIF_RSP_ERROR);
    return;
}

static void blk_check_ring(struct blkdev *blkdev)
{
    RING_IDX rc, rp;
    blkif_request_t req;

    if (!blkdev->rings.common.sring)
	return;

    for (;;) {
	blkdev->more_work = 0;
	rc = blkdev->rings.common.req_cons;
	rp = blkdev->rings.common.sring->req_prod;
	rmb(); /* Ensure we see queued requests up to 'rp'. */
	
	while ((rc != rp)) {
	    if (RING_REQUEST_CONS_OVERFLOW(&blkdev->rings.common, rc))
		break;
	    switch (blkdev->protocol) {
	    case BLKIF_PROTOCOL_NATIVE:
		memcpy(&req, RING_GET_REQUEST(&blkdev->rings.native, rc), sizeof(req));
		break;
	    case BLKIF_PROTOCOL_X86_32:
		blkif_get_x86_32_req(&req, RING_GET_REQUEST(&blkdev->rings.x86_32, rc));
		break;
	    case BLKIF_PROTOCOL_X86_64:
		blkif_get_x86_64_req(&req, RING_GET_REQUEST(&blkdev->rings.x86_64, rc));
		break;
	    }
	    blkdev->rings.common.req_cons = ++rc;
	    blk_handle_request(blkdev, &req);
	}
	if (!blkdev->more_work)
	    break;
    }
}

static int blk_alloc(struct xendev *xendev)
{
    struct blkdev *blkdev = container_of(xendev, struct blkdev, xendev);

    d1printf("%s: %p\n", __FUNCTION__, blkdev);
    blkdev->file = -1;
    return 0;
}

static int blk_setup(struct xendev *xendev)
{
    struct blkdev *blkdev = container_of(xendev, struct blkdev, xendev);

    d1printf("%s: %p\n", __FUNCTION__, blkdev);
    blk_setup_backend(blkdev);
    blk_connect_frontend(blkdev);
    blk_check_ring(blkdev);
    return 0;
}

static int blk_backend(struct xendev *xendev, char *node, char *val)
{
    struct blkdev *blkdev = container_of(xendev, struct blkdev, xendev);

    d2printf("%s: %p \"%s\" = \"%s\"\n", __FUNCTION__, blkdev, node,
	     val ? val : "<deleted>");
    blk_setup_backend(blkdev);
    return 0;
}

static int blk_frontend(struct xendev *xendev, char *node, char *val)
{
    struct blkdev *blkdev = container_of(xendev, struct blkdev, xendev);

    d2printf("%s: %p \"%s\" = \"%s\"\n", __FUNCTION__, blkdev, node,
	     val ? val : "<deleted>");
    blk_connect_frontend(blkdev);
    blk_check_ring(blkdev);
    return 0;
}

static int blk_event(struct xendev *xendev)
{
    struct blkdev *blkdev = container_of(xendev, struct blkdev, xendev);

    blk_check_ring(blkdev);
    return 0;
}

static int blk_free(struct xendev *xendev)
{
    struct blkdev *blkdev = container_of(xendev, struct blkdev, xendev);

    d1printf("%s: %p\n", __FUNCTION__, blkdev);

    if (-1 != blkdev->file)
	close(blkdev->file);
    if (blkdev->params)
	free(blkdev->params);
    if (blkdev->mode)
	free(blkdev->mode);

    vmcore_unmap(&blkdev->xendev);
    return 0;
}

static struct devops blkdev_ops = {
    .size  = sizeof(struct blkdev),
    .alloc = blk_alloc,
    .setup = blk_setup,
    .xs_be = blk_backend,
    .xs_fe = blk_frontend,
    .event = blk_event,
    .free  = blk_free,
};

/* ------------------------------------------------------------- */

static void usage(FILE *fp)
{
    fprintf(fp,
	    "\n"
	    "blkbackd -- no usage yet, sorry\n"
	    "\n");
}

int main(int argc, char *argv[])
{
    struct sigaction act,old;
    int maxfd, fd;
    fd_set set;
    int c;
                
    struct xs_handle *xs = NULL;

    for (;;) {
        if (-1 == (c = getopt(argc, argv, "hdp:")))
            break;
        switch (c) {
        case 'd':
	    debug++;
	    break;
	case 'p':
	    pidfile = optarg;
	    break;

        case 'h':
            usage(stdout);
            exit(0);
        default:
            usage(stderr);
            exit(1);
        }
    }

    /* setup signal handler */
    memset(&act,0,sizeof(act));
    sigemptyset(&act.sa_mask);
    act.sa_handler = catchsig;
    sigaction(SIGTERM,&act,&old);
    sigaction(SIGINT,&act,&old);

    evtchnd = xenner_evtchn_open();
    if (evtchnd < 0) {
	fprintf(stderr, "can't connect to evtchnd\n");
	exit(1);
    }

    xs = xs_daemon_open();
    if (!xs) {
	fprintf(stderr, "can't connect to xenstored\n");
	exit(1);
    }

    /* fork into background, handle pidfile */
    daemonize();

    xenstore_scan(xs, DEVPATH, &blkdev_ops);

    /* main loop */
    for (;;) {
	if (termsig)
	    break;
	
	FD_ZERO(&set);
	maxfd = 0;

	fd = xs_fileno(xs);
	FD_SET(fd, &set);
	if (maxfd < fd)
	    maxfd = fd;

	fd = xenner_evtchn_fd(evtchnd);
	FD_SET(fd, &set);
	if (maxfd < fd)
	    maxfd = fd;

	switch (select(maxfd+1, &set, NULL, NULL, NULL)) {
	case -1:
	    if (EINTR == errno)
		continue; /* termsig check */
	    perror("select");
	    break;
	case 0:
	    break;
	default:
	    break;
	}

	if (FD_ISSET(xs_fileno(xs), &set))
	    xenstore_update(xs);
	if (FD_ISSET(xenner_evtchn_fd(evtchnd), &set))
	    event_received();
    }
    return 0;
}
