#include <inttypes.h>
#include <xen/xen.h>

#include "emu32.h"

/* --------------------------------------------------------------------- */

static int map_next, map_max = 16;

/* --------------------------------------------------------------------- */

void *map_32(uint32_t maddr)
{
    uint32_t mfn = addr_to_frame(maddr);
    uint32_t off = addr_offset(maddr);
    uint32_t va;
    int i, flush = 0;

    for (i = 0; i < map_max; i++) {
	if (!test_pgflag_32(maps_32[i], _PAGE_PRESENT))
	    continue;
	if (get_pgframe_32(maps_32[i]) == mfn) {
	    /* cache hit */
	    goto found;
	}
    }

    i = map_next++ % map_max;
    if (test_pgflag_32(maps_32[i], _PAGE_PRESENT))
	flush = 1;
    maps_32[i] = get_pgentry_32(mfn, EMU_PGFLAGS);

found:
    va = XEN_MAP_32 + i*PAGE_SIZE;
    flush_tlb_addr(va);
    return (void*)va + off;
}

void *map_pae(uint64_t maddr)
{
    uint32_t mfn = addr_to_frame(maddr);
    uint32_t off = addr_offset(maddr);
    uint32_t va;
    int i, flush = 0;

    for (i = 0; i < map_max; i++) {
	if (!test_pgflag_pae(maps_pae[i], _PAGE_PRESENT))
	    continue;
	if (get_pgframe_pae(maps_pae[i]) == mfn) {
	    /* cache hit */
	    goto found;
	}
    }

    i = map_next++ % map_max;
    if (test_pgflag_pae(maps_pae[i], _PAGE_PRESENT))
	flush = 1;
    maps_pae[i] = get_pgentry_pae(mfn, EMU_PGFLAGS);

found:
    va = XEN_MAP_PAE + i*PAGE_SIZE;
    flush_tlb_addr(va);
    return (void*)va + off;
}

/* --------------------------------------------------------------------- */

uint32_t *find_pte_32(uint32_t va)
{
    uint32_t *lpt_base = (void*)XEN_LPT_32;
    uint32_t offset = va >> PAGE_SHIFT;

    return lpt_base + offset;
}

uint64_t *find_pte_pae(uint32_t va)
{
    uint64_t *lpt_base = (void*)XEN_LPT_PAE;
    uint32_t offset = va >> PAGE_SHIFT;

    return lpt_base + offset;
}

/* --------------------------------------------------------------------- */

void update_emu_mappings_32(uint32_t cr3_mfn)
{
    uint32_t *old_pgd;
    uint32_t *new_pgd;
    int idx;

    new_pgd  = map_32(frame_to_addr(cr3_mfn));

    idx = PGD_INDEX_32(XEN_M2P_32);
    if (!test_pgflag_32(new_pgd[idx], _PAGE_PRESENT)) {
	/* new one, must init static mappings, copy from current */
	old_pgd  = map_32(read_cr3() & PAGE_MASK);
	while (idx < PGD_COUNT_32) {
	    new_pgd[idx] = old_pgd[idx];
	    idx++;
	}
    }

    /* linear pgtable mapping */
    idx = PGD_INDEX_32(XEN_LPT_32);
    new_pgd[idx] = get_pgentry_32(cr3_mfn, LPT_PGFLAGS);
}

void update_emu_mappings_pae(uint32_t cr3_mfn)
{
    uint64_t *old_pgd, *old_pmd3;
    uint64_t *new_pgd, *new_pmd3;
    uint32_t mfn;
    int idx, i;

    new_pgd  = map_pae(frame_to_addr(cr3_mfn));
    new_pmd3 = map_pae(frame_to_addr(get_pgframe_pae(new_pgd[3])));

    idx = PMD_INDEX_PAE(XEN_M2P_PAE);
    if (!test_pgflag_pae(new_pmd3[idx], _PAGE_PRESENT)) {
	/* new one, must init static mappings, copy from current */
	old_pgd  = map_pae(read_cr3() & PAGE_MASK);
	old_pmd3 = map_pae(frame_to_addr(get_pgframe_pae(old_pgd[3])));
	while (idx < PMD_COUNT_PAE) {
	    new_pmd3[idx] = old_pmd3[idx];
	    idx++;
	}
    }

    /* linear pgtable mappings */
    idx = PMD_INDEX_PAE(XEN_LPT_PAE);
    for (i = 0; i < 4; i++) {
	if (test_pgflag_pae(new_pgd[i], _PAGE_PRESENT)) {
	    mfn = get_pgframe_pae(new_pgd[i]);
	    new_pmd3[idx+i] = get_pgentry_pae(mfn, LPT_PGFLAGS);
	} else {
	    new_pmd3[idx+i] = 0;
	}
    }
}
